from __future__ import annotations

import random
import math

from pytactoe.tictactoe import Piece, TicTacToe

MAX_EVAL = 10

class Engine:
    ''' Tic Tac Toe AI.'''
    @staticmethod
    def minimax_evaluate(game: TicTacToe, maximizing=True, depth=3) -> int:
        ''' Return minmax evaluation. '''
        if game.players != 2:
            raise ValueError('Minimax evaluation can only be applied to a 2-player game.')
        if game.winned:
            return +MAX_EVAL if game.winner == Piece(1) else -MAX_EVAL
        elif game.draw:
            return 0
        elif depth == 0:
            return 0

        if maximizing:
            value = -MAX_EVAL
            for move in game.legalmoves():
                game.move(*move)
                value = max(value, Engine.minimax_evaluate(game, False, depth-1))
                game.pop()
            return value

        else:
            value = +MAX_EVAL
            for move in game.legalmoves():
                game.move(*move)
                value = min(value, Engine.minimax_evaluate(game, True, depth-1))
                game.pop()
            return value

    @staticmethod
    def minimax_response(game: TicTacToe, depth=3) -> tuple:
        max_evaluation = -MAX_EVAL if game.turn == Piece(1) else +MAX_EVAL
        sgn = math.copysign(1, -max_evaluation) # Why is there no math.sign?
        best_move = None
        for move in game.legalmoves():
            game.move(*move)
            if Engine.minimax_evaluate(game, depth)*sgn >= max_evaluation*sgn:
                best_move = move
            game.pop()
        return best_move

    @staticmethod
    def random_response(game: 'TicTacToe', mode='random') -> tuple:
        ''' Return the best move on board. '''
        if game.ended:
            raise TypeError('Cannot response to ended game.')
        if mode == 'random':
            return random.choice(game.legalmoves())